﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Windows.Forms;
using System.Runtime.InteropServices;
using System.Diagnostics;

using Radar.Properties;

namespace Radar
{
	public partial class MainForm : Form
	{
		//bool m_bConnected;
		RadarConnection m_conn = null;
		bool m_bRunning = false;
		int? m_iFreq;

		public enum StatIndex
		{
			Average,
			Min,
			Max,
			SDNeg,
			SDPos,
			First,
			Second,
			Third,
			Count
		}
		
		ZedGraph.PointPairList m_pplFreqCount = new ZedGraph.PointPairList();
		ZedGraph.PointPairList m_pplFreqCountWiFi = new ZedGraph.PointPairList();
		ZedGraph.PointPairList[] m_pplFreqRSSI = new ZedGraph.PointPairList[(int)StatIndex.Count];
		ZedGraph.PointPairList m_pplFreqRSSIWiFi = new ZedGraph.PointPairList();
		ZedGraph.PointPairList[] m_pplFreqWidth = new ZedGraph.PointPairList[(int)StatIndex.Count];
		ZedGraph.PointPairList[] m_pplFreqTime = new ZedGraph.PointPairList[(int)StatIndex.Count];

		ZedGraph.PointPairList m_pplTimeCount = new ZedGraph.PointPairList();
		ZedGraph.PointPairList m_pplTimeCountWiFi = new ZedGraph.PointPairList();
		ZedGraph.PointPairList[] m_pplTimeRSSI = new ZedGraph.PointPairList[(int)StatIndex.Count];
		ZedGraph.PointPairList m_pplTimeRSSIWiFi = new ZedGraph.PointPairList();
		ZedGraph.PointPairList[] m_pplTimeWidth = new ZedGraph.PointPairList[(int)StatIndex.Count];
		ZedGraph.PointPairList[] m_pplTimeTime = new ZedGraph.PointPairList[(int)StatIndex.Count];

		ZedGraph.PointPairList[][] m_listLinePPLArrays;

		ZedGraph.PointPairList m_pplHistoWidth = new ZedGraph.PointPairList();
		ZedGraph.PointPairList m_pplHistoRSSI = new ZedGraph.PointPairList();
		ZedGraph.PointPairList m_pplHistoTime = new ZedGraph.PointPairList();

		/*ZedGraph.GraphPane[] m_panesAll;
		ZedGraph.GraphPane[] m_panesHisto;
		ZedGraph.GraphPane[] m_panesLine;*/

		ZedGraph.ZedGraphControl[] m_zgcAll;
		ZedGraph.ZedGraphControl[] m_zgcHisto;
		ZedGraph.ZedGraphControl[] m_zgcLines;

		List<ZedGraph.PointPairList> m_listLinePPLs = new List<ZedGraph.PointPairList>();

		Bitmap m_bmpGradient = null;

		public MainForm()
		{
			InitializeComponent();

			try
			{
				m_bmpGradient = (Bitmap)Image.FromFile("gradient.png");
			}
			catch
			{
			}

			m_listLinePPLArrays = new ZedGraph.PointPairList[] []{
				m_pplFreqRSSI,
				m_pplFreqWidth,
				m_pplFreqTime,

				m_pplTimeRSSI,
				m_pplTimeWidth,
				m_pplTimeTime
			};

			foreach (ZedGraph.PointPairList[] ppl in m_listLinePPLArrays)
			{
				for (int i = 0; i < (int)StatIndex.Count; ++i)
					ppl[i] = new ZedGraph.PointPairList();
			}

			/*m_panesAll = new ZedGraph.GraphPane[]{
				zgcFrequencyCount.GraphPane,
				zgcFrequencyRSSI.GraphPane,
				zgcFrequencyWidth.GraphPane,
				zgcFrequencyTime.GraphPane,

				zgcTimeCount.GraphPane,
				zgcTimeRSSI.GraphPane,
				zgcTimeWidth.GraphPane,
				zgcTimeTime.GraphPane,

				zgcHistoWidth.GraphPane,
				zgcHistoRSSI.GraphPane,
				zgcHistoTime.GraphPane
			};

			m_panesHisto = new ZedGraph.GraphPane[]{
				zgcHistoWidth.GraphPane,
				zgcHistoRSSI.GraphPane,
				zgcHistoTime.GraphPane
			};

			m_panesLine = new ZedGraph.GraphPane[]{
				zgcFrequencyCount.GraphPane,
				zgcFrequencyRSSI.GraphPane,
				zgcFrequencyWidth.GraphPane,
				zgcFrequencyTime.GraphPane,

				zgcTimeCount.GraphPane,
				zgcTimeRSSI.GraphPane,
				zgcTimeWidth.GraphPane,
				zgcTimeTime.GraphPane,
			};*/

			m_zgcAll = new ZedGraph.ZedGraphControl[]{
				zgcFrequencyCount,
				zgcFrequencyRSSI,
				zgcFrequencyWidth,
				zgcFrequencyTime,

				zgcTimeCount,
				zgcTimeRSSI,
				zgcTimeWidth,
				zgcTimeTime,

				zgcHistoWidth,
				zgcHistoRSSI,
				zgcHistoTime
			};

			m_zgcHisto = new ZedGraph.ZedGraphControl[]{
				zgcHistoWidth,
				zgcHistoRSSI,
				zgcHistoTime
			};

			m_zgcLines = new ZedGraph.ZedGraphControl[]{
				zgcFrequencyCount,
				zgcFrequencyRSSI,
				zgcFrequencyWidth,
				zgcFrequencyTime,

				zgcTimeCount,
				zgcTimeRSSI,
				zgcTimeWidth,
				zgcTimeTime
			};



			m_listLinePPLs.Add(m_pplFreqCount);
			m_listLinePPLs.Add(m_pplFreqCountWiFi);
			m_listLinePPLs.AddRange(m_pplFreqRSSI.ToList());
			m_listLinePPLs.Add(m_pplFreqRSSIWiFi);
			m_listLinePPLs.AddRange(m_pplFreqWidth.ToList());
			m_listLinePPLs.AddRange(m_pplFreqTime.ToList());

			m_listLinePPLs.Add(m_pplTimeCount);
			m_listLinePPLs.Add(m_pplTimeCountWiFi);
			m_listLinePPLs.AddRange(m_pplTimeRSSI.ToList());
			m_listLinePPLs.Add(m_pplTimeRSSIWiFi);
			m_listLinePPLs.AddRange(m_pplTimeWidth.ToList());
			m_listLinePPLs.AddRange(m_pplTimeTime.ToList());
		}

		List<ZedGraph.LineItem> m_listMultiCurves = new List<ZedGraph.LineItem>();

		List<ZedGraph.LineItem> SetupCurves(ZedGraph.ZedGraphControl zgc, ZedGraph.PointPairList[] ppls)
		{
			List<ZedGraph.LineItem> list = new List<ZedGraph.LineItem>();

			string[] strNames = Enum.GetNames(typeof(StatIndex));

			ZedGraph.ColorSymbolRotator colour = new ZedGraph.ColorSymbolRotator();

			//foreach (string str in strNames)
			for (int i = 0; i < (int)StatIndex.Count; ++i)
			{
				ZedGraph.LineItem li = zgc.GraphPane.AddCurve(/*str*/strNames[list.Count], ppls[list.Count], colour.NextColor);

				switch ((StatIndex)i)
				{
					case StatIndex.First:
					case StatIndex.Second:
					case StatIndex.Third:
						li.Line.Style = System.Drawing.Drawing2D.DashStyle.Dot;
						break;
					case StatIndex.Min:
					case StatIndex.Max:
						//li.Line.Width *= 2;
						break;
				}

				list.Add(li);
			}

			return list;
		}

		private void MainForm_Load(object sender, EventArgs e)
		{
			TehLib.Utils.LoadWindowState(this);

			txtFreqStep_TextChanged(this, null);

			UpdateParameterText();

			txtTimeQuantise.Text = Settings.Default.TimeQuantisation.ToString();

			int iFreq;
			if (int.TryParse(txtFreqStart.Text, out iFreq) && (iFreq >= nudFreq.Minimum) && (iFreq <= nudFreq.Maximum))
				nudFreq.Value = iFreq;

			////////////////////

			m_conn = new RadarConnection();
			m_conn.DispatchForm = this;
			m_conn.OnConnectionStateChange += m_conn_OnConnectionStateChange;
			m_conn.OnNewMessages += m_conn_OnNewMessages;

			////////////////////

			//foreach (ZedGraph.GraphPane gp in m_panesAll)
			foreach (ZedGraph.ZedGraphControl zgc in m_zgcAll)
			{
				ZedGraph.GraphPane gp = zgc.GraphPane;

				gp.Title.IsVisible = false;
				gp.XAxis.Title.IsVisible = false;
				gp.YAxis.Title.IsVisible = false;
				//gp.Legend.IsVisible = false;
				gp.Legend.FontSpec.Size = Settings.Default.FontSize + 1;
				gp.XAxis.Scale.FontSpec.Size = Settings.Default.FontSize;
				gp.YAxis.Scale.FontSpec.Size = Settings.Default.FontSize;
			}

			zgcHistoTime.GraphPane.XAxis.Title.IsVisible = true;
			zgcHistoTime.GraphPane.XAxis.Title.Text = "Time Difference";
			zgcHistoTime.GraphPane.XAxis.Title.FontSpec.Size = Settings.Default.FontSize + 1;
			zgcHistoTime.GraphPane.YAxis.Title.IsVisible = true;
			zgcHistoTime.GraphPane.YAxis.Title.Text = "Count";
			zgcHistoTime.GraphPane.YAxis.Title.FontSpec.Size = Settings.Default.FontSize + 1;

			foreach (ZedGraph.ZedGraphControl zgc in new ZedGraph.ZedGraphControl[] { zgcFrequencyTime, zgcTimeTime })
			{
				//zgc.GraphPane.XAxis.Title.IsVisible = true;
				zgc.GraphPane.YAxis.Title.IsVisible = true;
			}

			List<ZedGraph.LineItem> listLines = new List<ZedGraph.LineItem>();

			ZedGraph.LineItem liFreqCount = zgcFrequencyCount.GraphPane.AddCurve("Radar", m_pplFreqCount, Color.Red);
			//liFreqCount.Symbol.Type = ZedGraph.SymbolType.Default;
			listLines.Add(liFreqCount);

			ZedGraph.LineItem liFreqCountWiFi = zgcFrequencyCount.GraphPane.AddCurve("WiFi", m_pplFreqCountWiFi, Color.Green);
			listLines.Add(liFreqCountWiFi);

			ZedGraph.LineItem liFreqRSSIWiFi = zgcFrequencyRSSI.GraphPane.AddCurve("WiFi", m_pplFreqRSSIWiFi, Color.LightGreen);
			liFreqRSSIWiFi.Line.Width *= 2;
			liFreqRSSIWiFi.Symbol.Type = ZedGraph.SymbolType.Circle;
			listLines.Add(liFreqRSSIWiFi);

			//ZedGraph.LineItem liFreqRSSI = zgcFrequencyRSSI.GraphPane.AddCurve("", m_pplFreqRSSI, Color.Blue);
			m_listMultiCurves.AddRange(SetupCurves(zgcFrequencyRSSI, m_pplFreqRSSI));
			m_listMultiCurves.AddRange(SetupCurves(zgcFrequencyWidth, m_pplFreqWidth));
			m_listMultiCurves.AddRange(SetupCurves(zgcFrequencyTime, m_pplFreqTime));

			ZedGraph.LineItem liTimeCount = zgcTimeCount.GraphPane.AddCurve("Radar", m_pplTimeCount, Color.Red);
			listLines.Add(liTimeCount);

			ZedGraph.LineItem liTimeCountWiFi = zgcTimeCount.GraphPane.AddCurve("WiFi", m_pplTimeCountWiFi, Color.Green);
			listLines.Add(liTimeCountWiFi);

			ZedGraph.LineItem liTimeRSSIWiFi = zgcTimeRSSI.GraphPane.AddCurve("WiFi", m_pplTimeRSSIWiFi, Color.LightGreen);
			liTimeRSSIWiFi.Line.Width *= 2;
			liTimeRSSIWiFi.Symbol.Type = ZedGraph.SymbolType.Circle;
			listLines.Add(liTimeRSSIWiFi);

			//ZedGraph.LineItem liTimeRSSI = zgcTimeRSSI.GraphPane.AddCurve("", m_pplTimeRSSI, Color.Blue);
			m_listMultiCurves.AddRange(SetupCurves(zgcTimeRSSI, m_pplTimeRSSI));
			m_listMultiCurves.AddRange(SetupCurves(zgcTimeWidth, m_pplTimeWidth));
			m_listMultiCurves.AddRange(SetupCurves(zgcTimeTime, m_pplTimeTime));

			listLines.AddRange(m_listMultiCurves);

			//foreach (ZedGraph.LineItem li in new ZedGraph.LineItem[] { liFreqCount, liFreqRSSI, liTimeCount, liTimeRSSI })
			foreach (ZedGraph.LineItem li in listLines)
			{
				li.Symbol.Size = 2;
			}

			//foreach (ZedGraph.GraphPane gp in m_panesLine)
			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
			{
				zgc.GraphPane.YAxis.Scale.Min = 0;
			}

			zgcHistoTime.GraphPane.XAxis.Scale.Min = 0;

			string[] strHistoLabels = new string[256];
			for (int i = 0; i < 256; i++)
			{
				m_pplHistoWidth.Add(i, 0);
				m_pplHistoRSSI.Add(i, 0);
				strHistoLabels[i] = i.ToString();
			}

			foreach (ZedGraph.GraphPane gp in new ZedGraph.GraphPane[] { zgcHistoWidth.GraphPane, zgcHistoRSSI.GraphPane })
			{
				gp.XAxis.Scale.Min = 0;
				gp.XAxis.Scale.Mag = 255;
				gp.YAxis.Scale.Min = 0;
				gp.XAxis.Type = ZedGraph.AxisType.Text;
				gp.XAxis.Scale.TextLabels = strHistoLabels;
				//zgcWidths.GraphPane.YAxis
			}

			ZedGraph.BarItem biWidths = zgcHistoWidth.GraphPane.AddBar("", m_pplHistoWidth, Color.Blue);
			//biWidths
			zgcHistoWidth.AxisChange();

			ZedGraph.BarItem biRSSI = zgcHistoRSSI.GraphPane.AddBar("", m_pplHistoRSSI, Color.Red);
			//biWidths
			zgcHistoRSSI.AxisChange();

			ZedGraph.BarItem biTime = zgcHistoTime.GraphPane.AddBar("", m_pplHistoTime, Color.Green);
			//biTime
			zgcHistoTime.AxisChange();

			////////////////////

			tcGraphs.SelectedIndex = Settings.Default.MajorTab;
			switch (tcGraphs.SelectedIndex)
			{
				case 0:
					tcFrequencyGraphs.SelectedIndex = Settings.Default.MinorTab;
					break;
				case 1:
					tcTimeGraphs.SelectedIndex = Settings.Default.MinorTab;
					break;
				case 2:
					tcHistoGraphs.SelectedIndex = Settings.Default.MinorTab;
					break;
			}
		}

		void m_conn_OnNewMessages(object sender, MessageList e)
		{
			foreach (string str in e.List)
			{
				int iIndex = str.IndexOf(" ");
				string strCmd;
				if (iIndex > -1)
					strCmd = str.Substring(0, iIndex);
				else
					strCmd = str;

				if (strCmd == "END")
				{
					Stop(false);
				}
				else if (strCmd == "FREQ")
				{
					SetFrequency(int.Parse(str.Substring(iIndex + 1)));

					tsslStatus.Text = string.Format("New frequency set: {0}", m_iFreq);
				}
				else if (strCmd == "DATA")
				{
					if (m_bRunning == false)
						continue;

					byte[] data = null;

					int iIndex2 = str.IndexOf(" ", iIndex + 1);
					string strData = string.Empty;
					if (iIndex2 > -1)
						strData = str.Substring(iIndex2 + 1).Trim();
					else
						iIndex2 = str.Length;

					try
					{
						data = Convert.FromBase64String(strData);

						SetFrequency(int.Parse(str.Substring(iIndex + 1, iIndex2 - (iIndex + 1))));

						ProcessData(data);
					}
					catch (Exception ex)
					{
						Debug.WriteLine(string.Format("{0} \"{1}\": {2}", strData.Length, strData, ex));

						tsslStatus.Text = "Processing error: " + ex.Message;
					}
				}
			}
		}

		void SetFrequency(int iFreq)
		{
			m_iFreq = iFreq;

			txtFreq.Text = iFreq.ToString();
		}

		void ProcessData(byte[] data)
		{
			//Debug.WriteLine(string.Format("Processing data for frequency {0}: {1} items, {2} bytes", m_iFreq, data.Length / 8, data.Length));

			const int iSize = 8;

			uint iOverflows = 0;
			uint? nTime = null;
			byte[] buffer = new byte[iSize];
			List<RadarError> list = new List<RadarError>();
			//List<RadarError> listRE = new List<RadarError>();
			List<uint> listOverflows = new List<uint>();

			for (int i = 0; i < data.Length; i += 8)
			{
				Array.Copy(data, i, buffer, 0, iSize);
				GCHandle handle = GCHandle.Alloc(buffer, GCHandleType.Pinned);
				RadarError re = (RadarError)Marshal.PtrToStructure(handle.AddrOfPinnedObject(), typeof(RadarError));
				handle.Free();

				list.Add(re);

				if ((nTime.HasValue) && (re.TSF < nTime.Value))
					++iOverflows;
				listOverflows.Add(iOverflows);

				nTime = re.TSF;
			}

			const uint nOverflowAmount = (0x7fff + 1);
			uint nOverflows = 0;
			if (listOverflows.Count > 0)
				nOverflows = listOverflows[listOverflows.Count - 1];

			for (int i = 0; i < list.Count; ++i)
			{
				RadarError re = list[i];

				re.TSF -= (nOverflowAmount * (nOverflows - listOverflows[i]));

				list[i] = re;

				//Debug.Assert(re.TSF <= nTime.Value);	// Happens at HW counter-rollover

				//if (i > 0)
				//	Debug.Assert(re.TSF > list[i - 1].TSF);
			}

			ProcessRadarErrors(list);
		}

		////////////////////////////////////////////////////////////////////////////////////////

		void ResetHistograms()
		{
			for (int i = 0; i < 256; i++)
			{
				m_pplHistoWidth[i].Y = 0;
				m_pplHistoRSSI[i].Y = 0;
			}

			m_pplHistoTime.Clear();

			//zgcWidths.AxisChange();
			/*zgcHistoWidth.Invalidate();
			zgcHistoRSSI.Invalidate();
			zgcHistoTime.Invalidate();*/

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcHisto)
				zgc.Invalidate();
		}

		void ResetLineGraphs()
		{
			foreach (ZedGraph.PointPairList ppl in m_listLinePPLs)
				ppl.Clear();

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();

			/*zgcFrequencyCount.Invalidate();
			zgcFrequencyRSSI.Invalidate();
			zgcFrequencyTime.Invalidate();

			zgcTimeCount.Invalidate();
			zgcTimeRSSI.Invalidate();
			zgcTimeTime.Invalidate();*/
		}

		string GetHistoItem(List<KeyValuePair<double, uint>> list, int i)
		{
			if (i > (list.Count - 1))
				return "-";

			KeyValuePair<double, uint> kv = list[i];

			return string.Format("{0}\t({1} times)", kv.Key, kv.Value);
		}

		void AddToLine(double d, Statistics stats, ZedGraph.PointPairList[] ppl)
		{
			ppl[0].Add(d, stats.Average);
			ppl[1].Add(d, stats.Min);
			ppl[2].Add(d, stats.Max);
			ppl[3].Add(d, stats.Average - stats.StandardDeviation);
			ppl[4].Add(d, stats.Average + stats.StandardDeviation);

			List<KeyValuePair<double,uint>> list = stats.GetHistogram(Statistics.HistogramSorting.CountDesc);

			ppl[5].Add(d, (list.Count > 0 ? list[0].Key : 0));
			ppl[6].Add(d, (list.Count > 1 ? list[1].Key : 0));
			ppl[7].Add(d, (list.Count > 2 ? list[2].Key : 0));
		}

		List<RadarError> m_listLastErrors = new List<RadarError>();

		void ProcessRadarErrors(List<RadarError> list)
		{
			ResetHistograms();

			Statistics statsRSSI = new Statistics(true);
			Statistics statsRSSIWiFi = new Statistics();
			Statistics statsWidth = new Statistics(true);
			Statistics statsTime = new Statistics(true);

			if (chkAccumulateTime.Checked == false)
				m_listLastErrors.Clear();

			//int i = 0;
			uint? nLastTime = null;
			//foreach (RadarError re in list)
			uint nWiFi = 0, nRadar = 0;
			for (int i = 0; i < list.Count; ++i)
			{
				RadarError re = list[i];

				if (re.Type == 0)
				{
					m_pplHistoWidth[re.Width].Y += 1;
					m_pplHistoRSSI[re.RSSI].Y += 1;

					statsRSSI.Add(re.RSSI);
					statsWidth.Add(re.Width);

					if (nLastTime.HasValue)
					{
						//Debug.Assert(re.TSF > nLastTime.Value);

						uint nTimeDiff = re.TSF - nLastTime.Value;
						statsTime.Add(nTimeDiff);
					}

					m_listLastErrors.Add(re);

					nLastTime = re.TSF;

					++nRadar;
				}
				else if (re.Type == 1)
				{
					statsRSSIWiFi.Add(re.RSSI);

					++nWiFi;
				}

				//++i;
			}

			m_pplFreqCount.Add(m_iFreq.Value, nRadar);
			m_pplFreqCountWiFi.Add(m_iFreq.Value, nWiFi);
			//m_pplFreqRSSI.Add(m_iFreq.Value, statsRSSI.Average);
			AddToLine(m_iFreq.Value, statsRSSI, m_pplFreqRSSI);
			m_pplFreqRSSIWiFi.Add(m_iFreq.Value, statsRSSIWiFi.Max);
			AddToLine(m_iFreq.Value, statsWidth, m_pplFreqWidth);
			AddToLine(m_iFreq.Value, statsTime, m_pplFreqTime);

			int iX = m_pplTimeCount.Count;
			m_pplTimeCount.Add(iX, nRadar);
			m_pplTimeCountWiFi.Add(iX, nWiFi);
			//m_pplTimeRSSI.Add(iX, statsRSSI.Average);
			AddToLine(iX, statsRSSI, m_pplTimeRSSI);
			m_pplTimeRSSIWiFi.Add(iX, statsRSSIWiFi.Max);
			AddToLine(iX, statsWidth, m_pplTimeWidth);
			AddToLine(iX, statsTime, m_pplTimeTime);

			uint nMaxTimeCount = 0;

			List<KeyValuePair<double, uint>> listTimeHisto = statsTime.GetHistogram(Statistics.HistogramSorting.CountDesc);
			if (chkQuantiseTime.Checked == false)
			{
				if (listTimeHisto.Count > 0)
				{
					foreach (KeyValuePair<double, uint> kv in listTimeHisto)
						m_pplHistoTime.Add(kv.Key, kv.Value);

					nMaxTimeCount = listTimeHisto[0].Value;
				}
			}
			else
			{
				List<KeyValuePair<double, uint>> listTimeHistoBA = statsTime.GetHistogram(Statistics.HistogramSorting.BinAsc);

				//List<string> listLabels = new List<string>();

				if (listTimeHistoBA.Count > 0)
				{
					double dStep = listTimeHistoBA[listTimeHistoBA.Count - 1].Key / (double)Settings.Default.TimeQuantisation;
					uint nBin = 0, nSum = 0;

					foreach (KeyValuePair<double, uint> kv in listTimeHistoBA)
					{
						uint n = (uint)Math.Floor(kv.Key / dStep);

						if (n != nBin)
						{
							m_pplHistoTime.Add(((int)(nBin * dStep)), nSum);
							//listLabels.Add(((int)(nBin * dStep)).ToString());
							nSum = 0;

							nBin = n;
						}

						nSum += kv.Value;

						if (nSum > nMaxTimeCount)
							nMaxTimeCount = nSum;
					}

					if (nSum > 0)
					{
						m_pplHistoTime.Add(((int)(nBin * dStep)), nSum);
						//listLabels.Add(((int)(nBin * dStep)).ToString());
					}
				}

				//zgcHistoTime.GraphPane.XAxis.Scale.TextLabels = listLabels.ToArray();
			}

			if (nMaxTimeCount > 0)
			{
				if ((m_iTimeHistogramMaxY.HasValue == false) || (nMaxTimeCount > m_iTimeHistogramMaxY.Value))
					m_iTimeHistogramMaxY = (int)nMaxTimeCount;

				zgcHistoTime.GraphPane.YAxis.Scale.Max = m_iTimeHistogramMaxY.Value;
			}

			if (chkTimeHistogramMaxXAxisScale.Checked)
			{
				if ((m_iTimeHistogramMaxX.HasValue == false) || (statsTime.Max > m_iTimeHistogramMaxX.Value))
					m_iTimeHistogramMaxX = (int)statsTime.Max;

				zgcHistoTime.GraphPane.XAxis.Scale.Max = m_iTimeHistogramMaxX.Value;
			}

			List<KeyValuePair<double, uint>> listWidthHisto = statsWidth.GetHistogram(Statistics.HistogramSorting.CountDesc);
			if (listWidthHisto.Count > 0)
			{
				if ((m_iWidthHistogramMaxY.HasValue == false) || (listWidthHisto[0].Value > m_iWidthHistogramMaxY.Value))
					m_iWidthHistogramMaxY = (int)listWidthHisto[0].Value;

				zgcHistoWidth.GraphPane.YAxis.Scale.Max = m_iWidthHistogramMaxY.Value;
			}

			List<KeyValuePair<double, uint>> listRSSIHisto = statsWidth.GetHistogram(Statistics.HistogramSorting.CountDesc);
			if (listRSSIHisto.Count > 0)
			{
				if ((m_iRSSIHistogramMaxY.HasValue == false) || (listRSSIHisto[0].Value > m_iRSSIHistogramMaxY.Value))
					m_iRSSIHistogramMaxY = (int)listRSSIHisto[0].Value;

				zgcHistoRSSI.GraphPane.YAxis.Scale.Max = m_iRSSIHistogramMaxY.Value;
			}

			uint nDiff = 0;
			if (list.Count > 0)
				nDiff = nLastTime.GetValueOrDefault(0) - list[0].TSF;

			txtTime.Text = string.Format(
				"Diff:\t{10}\n" +
				"Count:\t{0}\n" +
				"Min:\t{1}\n" +
				"Max:\t{2}\n" +
				"Ave:\t{3}\n" +
				"SD:\t{4}\n" +
				"1st:\t{5}\n" +
				"2nd:\t{6}\n" +
				"3rd:\t{7}\n" +
				"4th:\t{8}\n" +
				"5th:\t{9}\n",
				listTimeHisto.Count,
				statsTime.Min,
				statsTime.Max,
				statsTime.Average,
				statsTime.StandardDeviation,
				GetHistoItem(listTimeHisto, 0),
				GetHistoItem(listTimeHisto, 1),
				GetHistoItem(listTimeHisto, 2),
				GetHistoItem(listTimeHisto, 3),
				GetHistoItem(listTimeHisto, 4),
				nDiff
				).Replace("\n", "\r\n");

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcAll)
			{
				zgc.AxisChange();
				zgc.Invalidate();
			}

			pbTime.Invalidate();
			pbRSSI.Invalidate();

			tsslStatus.Text = string.Format("Processed {0} items ({1})", list.Count, DateTime.Now.ToLongTimeString());
		}

		////////////////////////////////////////////////////////////////////////////////////////

		void Stop(bool bManual)
		{
			if (bManual)
			{
				m_conn.Send("STOP");

				tsslStatus.Text = "Stopped";
			}
			else
				tsslStatus.Text = "Finished";

			m_iFreq = null;

			m_bRunning = false;

			btnStart.Text = "Start";

			txtFreqStart.Enabled = true;
			txtFreqStop.Enabled = chkFrequencyRangeStop.Checked;
			txtFreqStep.Enabled = true;
			txtInterval.Enabled = true;
			chkFrequencyRangeStop.Enabled = true;
		}

		void m_conn_OnConnectionStateChange(object sender, TehLib.ConnectionState e)
		{
			if (e.Connected)
			{
				btnConnect.Text = "Disconnect";
				btnConnect.Enabled = true;

				tsslStatus.Text = "Connected";

				btnStart.Enabled = true;

				SendAllParameters();
			}
			else
			{
				btnConnect.Text = "Connect";
				btnConnect.Enabled = true;

				btnStart.Enabled = false;

				tsslStatus.Text = "Disconnected";
			}
		}

		private void MainForm_FormClosing(object sender, FormClosingEventArgs e)
		{
			if (m_bmpGradient != null)
				m_bmpGradient.Dispose();

			Settings.Default.MajorTab = tcGraphs.SelectedIndex;
			switch (tcGraphs.SelectedIndex)
			{
				case 0:
					Settings.Default.MinorTab = tcFrequencyGraphs.SelectedIndex;
					break;
				case 1:
					Settings.Default.MinorTab = tcTimeGraphs.SelectedIndex;
					break;
				case 2:
					Settings.Default.MinorTab = tcHistoGraphs.SelectedIndex;
					break;
			}

			TehLib.Utils.SaveWindowState(this);
		}

		private void btnConnect_Click(object sender, EventArgs e)
		{
			if (m_conn.Connected == false)
			{
				if (m_conn.Connect(txtServer.Text) == false)
					MessageBox.Show("Failed to connect");
				else
				{
					btnConnect.Enabled = false;
					tsslStatus.Text = "Connecting...";
				}
			}
			else
			{
				Stop(true);

				m_conn.Disconnect();
			}
		}

		private void btnStart_Click(object sender, EventArgs e)
		{
			if (m_bRunning)
			{
				Stop(true);
			}
			else
			{
				if (m_conn.Connected == false)
					return;

				ResetLineGraphs();

				m_listLastErrors.Clear();

				//nudFreq.Minimum = int.Parse(txtFreqStart.Text);
				//nudFreq.Maximum = int.Parse(txtFreqStop.Text);
				//if (nudFreq.Value == 0)
				int i;
				if (int.TryParse(txtFreqStart.Text, out i) && (i >= nudFreq.Minimum) && (i <= nudFreq.Maximum))
				{
					m_bIgnoreNUDFreqChange = true;
					nudFreq.Value = i;
					m_bIgnoreNUDFreqChange = false;
				}

				m_conn.Send(string.Format("START {0} {1} {2} {3}",
					txtFreqStart.Text,
					(chkFrequencyRangeStop.Checked ? txtFreqStop.Text : txtFreqStart.Text),
					txtFreqStep.Text,
					txtInterval.Text));

				btnStart.Text = "Stop";

				txtFreqStart.Enabled = false;
				txtFreqStop.Enabled = false;
				txtFreqStep.Enabled = false;
				txtInterval.Enabled = false;
				chkFrequencyRangeStop.Enabled = false;

				m_bRunning = true;

				tsslStatus.Text = "Starting...";
			}
		}

		private void txtFreqStep_TextChanged(object sender, EventArgs e)
		{
			int i;
			if (int.TryParse(txtFreqStep.Text, out i))
				nudFreq.Increment = i;
		}

		bool m_bIgnoreNUDFreqChange = false;

		private void nudFreq_ValueChanged(object sender, EventArgs e)
		{
			if ((m_conn != null) && (m_bIgnoreNUDFreqChange == false))
			{
				if (m_conn.Send(string.Format("FREQ {0}", nudFreq.Value)))
					tsslStatus.Text = string.Format("Changing frequency to {0}...", nudFreq.Value);
			}
		}

		private void exitToolStripMenuItem_Click(object sender, EventArgs e)
		{
			Application.Exit();
		}

		private void tbFIRPower_Scroll(object sender, EventArgs e)
		{
			m_conn.Send(string.Format("FIRPWR {0}", tbFIRPower.Value));

			if (sender != this)
				UpdateParameterText();
		}

		private void tbPulseHeight_Scroll(object sender, EventArgs e)
		{
			m_conn.Send(string.Format("PHEIGHT {0}", tbPulseHeight.Value));

			if (sender != this)
				UpdateParameterText();
		}

		private void tbRadarRSSI_Scroll(object sender, EventArgs e)
		{
			m_conn.Send(string.Format("RSSI {0}", tbRadarRSSI.Value));

			if (sender != this)
				UpdateParameterText();
		}

		private void tbPulseRSSI_Scroll(object sender, EventArgs e)
		{
			m_conn.Send(string.Format("PRSSI {0}", tbPulseRSSI.Value));

			if (sender != this)
				UpdateParameterText();
		}

		private void tbInBandThreshold_Scroll(object sender, EventArgs e)
		{
			m_conn.Send(string.Format("INBAND {0}", tbInBandThreshold.Value));

			if (sender != this)
				UpdateParameterText();
		}

		bool m_bIgnoreParameterTextChanges = false;

		void UpdateParameterText()
		{
			m_bIgnoreParameterTextChanges = true;

			txtFIRPower.Text = tbFIRPower.Value.ToString();
			txtPulseHeight.Text = tbPulseHeight.Value.ToString();
			txtRadarRSSI.Text = tbRadarRSSI.Value.ToString();
			txtPulseRSSI.Text = tbPulseRSSI.Value.ToString();
			txtInBandThreshold.Text = tbInBandThreshold.Value.ToString();

			m_bIgnoreParameterTextChanges = false;
		}

		void SendAllParameters()
		{
			tbFIRPower_Scroll(this, null);
			tbPulseHeight_Scroll(this, null);
			tbRadarRSSI_Scroll(this, null);
			tbPulseRSSI_Scroll(this, null);
			tbInBandThreshold_Scroll(this, null);

			UpdateParameterText();
		}

		private void txtFIRPower_TextChanged(object sender, EventArgs e)
		{
			if (m_bIgnoreParameterTextChanges)
				return;

			int i;
			if (int.TryParse(txtFIRPower.Text, out i) && (i >= tbFIRPower.Minimum) && (i <= tbFIRPower.Maximum))
				tbFIRPower.Value = i;
		}

		private void txtPulseHeight_TextChanged(object sender, EventArgs e)
		{
			if (m_bIgnoreParameterTextChanges)
				return;

			int i;
			if (int.TryParse(txtPulseHeight.Text, out i) && (i >= tbPulseHeight.Minimum) && (i <= tbPulseHeight.Maximum))
				tbPulseHeight.Value = i;
		}

		private void txtRadarRSSI_TextChanged(object sender, EventArgs e)
		{
			if (m_bIgnoreParameterTextChanges)
				return;

			int i;
			if (int.TryParse(txtRadarRSSI.Text, out i) && (i >= tbRadarRSSI.Minimum) && (i <= tbRadarRSSI.Maximum))
				tbRadarRSSI.Value = i;
		}

		private void txtPulseRSSI_TextChanged(object sender, EventArgs e)
		{
			if (m_bIgnoreParameterTextChanges)
				return;

			int i;
			if (int.TryParse(txtPulseRSSI.Text, out i) && (i >= tbPulseRSSI.Minimum) && (i <= tbPulseRSSI.Maximum))
				tbPulseRSSI.Value = i;
		}

		private void txtInBandThreshold_TextChanged(object sender, EventArgs e)
		{
			if (m_bIgnoreParameterTextChanges)
				return;

			int i;
			if (int.TryParse(txtInBandThreshold.Text, out i) && (i >= tbInBandThreshold.Minimum) && (i <= tbInBandThreshold.Maximum))
				tbInBandThreshold.Value = i;
		}

		private void pbTime_Paint(object sender, PaintEventArgs e)
		{
			PaintTime(pbTime, PaintAttribute.Width, e);
		}

		private void pbRSSI_Paint(object sender, PaintEventArgs e)
		{
			PaintTime(pbRSSI, PaintAttribute.RSSI, e);
		}

		enum PaintAttribute
		{
			Width,
			RSSI
		}

		void PaintTime(PictureBox pb, PaintAttribute pa, PaintEventArgs e)
		{
			if (pb.Width <= 0 || pb.Height <= 0)
				return;

			if (m_listLastErrors == null)
				return;
			else if (m_listLastErrors.Count < 2)
				return;

			uint nOffset = m_listLastErrors[0].TSF;
			uint nDiff = m_listLastErrors[m_listLastErrors.Count - 1].TSF - nOffset;

			double dStep = pb.Width / (double)nDiff;

			RadarError? reLast = null;
			//double dOffset = 0;
			List<float> listSingle = (chkHighlightShortTimeDifferences.Checked ? new List<float>() : null);
			foreach (RadarError re in m_listLastErrors)
			{
				if (reLast.HasValue)
				{
					Color colour = Color.Black;

					Color colourFill = Color.White;
					if (pa == PaintAttribute.Width)
					{
						if (m_bmpGradient != null)
							colourFill = m_bmpGradient.GetPixel(reLast.Value.Width, 0);
						else
							colourFill = Color.FromArgb(0, 0, reLast.Value.Width);
					}
					else if (pa == PaintAttribute.RSSI)
					{
						if (m_bmpGradient != null)
							colourFill = m_bmpGradient.GetPixel(reLast.Value.RSSI, 0);
						else
							colourFill = Color.FromArgb(reLast.Value.RSSI, 0, 0);
					}

					Brush b = new SolidBrush(colourFill);
					Pen p = new Pen(colour, 1.0f);

					RectangleF rect = new RectangleF();
					//(float)dOffset, 0, (float)(dOffset + ((double)(re.TSF - reLast.Value.TSF) * dStep))
					rect.X = (float)((double)(reLast.Value.TSF - nOffset) * dStep);
					rect.Width = (float)((double)(re.TSF - reLast.Value.TSF) * dStep);
					rect.Y = 0;
					rect.Height = pb.Height - 1;

					e.Graphics.FillRectangle(b, rect);
					e.Graphics.DrawRectangle(p, rect.X, rect.Y, rect.Width, rect.Height);

					if (rect.Width < 2)
					{
						//e.Graphics.DrawLine(Pens.White, rect.X, 0, rect.X, rect.Height);

						if (listSingle != null)
							listSingle.Add(rect.X);
					}

					b.Dispose();
					p.Dispose();

					//dOffset += ((double)re.TSF * dStep);
				}

				reLast = re;
			}

			if (listSingle != null)
			{
				foreach (float f in listSingle)
					e.Graphics.DrawLine(Pens.White, f, 0, f, pb.Height - 1);
			}
		}

		private void chkQuantiseTime_CheckedChanged(object sender, EventArgs e)
		{
			if (chkQuantiseTime.Checked)
				btnSetTimeQuantisation_Click(this, null);
			else
				zgcHistoTime.GraphPane.XAxis.Scale.TextLabels = null;
		}

		private void btnSetTimeQuantisation_Click(object sender, EventArgs e)
		{
			uint nTimeQuantisation;
			if (uint.TryParse(txtTimeQuantise.Text, out nTimeQuantisation))
				Settings.Default.TimeQuantisation = nTimeQuantisation;
		}

		private void btnCopyFrequencyToInput_Click(object sender, EventArgs e)
		{
			//if (m_iFreq.HasValue)
			//	nudFreq.Value = m_iFreq.Value;

			int i;
			if (int.TryParse(txtFreq.Text, out i) && (i >= nudFreq.Minimum) && (i <= nudFreq.Maximum))
				nudFreq.Value = i;	// Will trigger callback
		}

		private void chkFrequencyRangeStop_CheckedChanged(object sender, EventArgs e)
		{
			txtFreqStop.Enabled = chkFrequencyRangeStop.Checked;
		}

		private void aboutToolStripMenuItem_Click(object sender, EventArgs e)
		{
			MessageBox.Show("By Balint Seeber (balint@spench.net) 2012\n\nhttp://spench.net/r/RADAR", "About", MessageBoxButtons.OK, MessageBoxIcon.Information);
		}

		private void defaultallToolStripMenuItem_Click(object sender, EventArgs e)
		{
			defaultFIRPowerToolStripMenuItem_Click(this, null);
			defaultpulseHeightToolStripMenuItem_Click(this, null);
			defaultRADARRSSIToolStripMenuItem_Click(this, null);
			defaultpulseRSSIToolStripMenuItem_Click(this, null);
			defaultinbandThresholdToolStripMenuItem_Click(this, null);

			SendAllParameters();
		}

		private void defaultFIRPowerToolStripMenuItem_Click(object sender, EventArgs e)
		{
			tbFIRPower.Value = -35;

			if (sender != this)
				tbFIRPower_Scroll(this, null);
		}

		private void defaultpulseHeightToolStripMenuItem_Click(object sender, EventArgs e)
		{
			tbPulseHeight.Value = 14;

			if (sender != this)
				tbPulseHeight_Scroll(this, null);
		}

		private void defaultRADARRSSIToolStripMenuItem_Click(object sender, EventArgs e)
		{
			tbRadarRSSI.Value = 20;

			if (sender != this)
				tbRadarRSSI_Scroll(this, null);
		}

		private void defaultpulseRSSIToolStripMenuItem_Click(object sender, EventArgs e)
		{
			tbPulseRSSI.Value = 6;

			if (sender != this)
				tbPulseRSSI_Scroll(this, null);
		}

		private void defaultinbandThresholdToolStripMenuItem_Click(object sender, EventArgs e)
		{
			tbInBandThreshold.Value = 4;

			if (sender != this)
				tbInBandThreshold_Scroll(this, null);
		}

		void ChangeCurveVisibility(StatIndex si, bool bVisible)
		{
			foreach (ZedGraph.LineItem li in m_listMultiCurves)
			{
				foreach (ZedGraph.PointPairList[] ppls in m_listLinePPLArrays)
				{
					if (ppls.Contains(li.Points))
					{
						if ((int)si == ppls.ToList().IndexOf(li.Points as ZedGraph.PointPairList))
							li.IsVisible = bVisible;
					}
				}
			}
		}

		private void chkShowAverage_CheckedChanged(object sender, EventArgs e)
		{
			ChangeCurveVisibility(StatIndex.Average, chkShowAverage.Checked);

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();
		}

		private void chkShowMin_CheckedChanged(object sender, EventArgs e)
		{
			ChangeCurveVisibility(StatIndex.Min, chkShowMin.Checked);

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();
		}

		private void chkShowStandardDeviation_CheckedChanged(object sender, EventArgs e)
		{
			ChangeCurveVisibility(StatIndex.SDNeg, chkShowStandardDeviation.Checked);
			ChangeCurveVisibility(StatIndex.SDPos, chkShowStandardDeviation.Checked);

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();
		}

		private void chkShowTop_CheckedChanged(object sender, EventArgs e)
		{
			ChangeCurveVisibility(StatIndex.First, chkShowTop.Checked);
			ChangeCurveVisibility(StatIndex.Second, chkShowTop.Checked);
			ChangeCurveVisibility(StatIndex.Third, chkShowTop.Checked);

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();
		}

		private void chkShowMax_CheckedChanged(object sender, EventArgs e)
		{
			ChangeCurveVisibility(StatIndex.Max, chkShowMax.Checked);

			foreach (ZedGraph.ZedGraphControl zgc in m_zgcLines)
				zgc.Invalidate();
		}

		int? m_iTimeHistogramMaxX = null;

		private void chkTimeHistogramMaxXAxisScale_CheckedChanged(object sender, EventArgs e)
		{
			if (chkTimeHistogramMaxXAxisScale.Checked == false)
			{
				zgcHistoTime.GraphPane.XAxis.Scale.MaxAuto = true;
				m_iTimeHistogramMaxX = null;
			}
		}

		int? m_iRSSIHistogramMaxY = null, m_iWidthHistogramMaxY = null, m_iTimeHistogramMaxY = null;

		private void resetYAxisScaleToolStripMenuItem_Click(object sender, EventArgs e)
		{
			zgcHistoWidth.GraphPane.YAxis.Scale.MaxAuto = true;
			m_iWidthHistogramMaxY = null;
			
			zgcHistoRSSI.GraphPane.YAxis.Scale.MaxAuto = true;
			m_iRSSIHistogramMaxY = null;

			zgcHistoTime.GraphPane.YAxis.Scale.MaxAuto = true;
			m_iTimeHistogramMaxY = null;
		}

		private void chkHighlightShortTimeDifferences_CheckedChanged(object sender, EventArgs e)
		{
			pbRSSI.Invalidate();
			pbTime.Invalidate();
		}

		private void btnClear_Click(object sender, EventArgs e)
		{
			foreach (ZedGraph.PointPairList ppl in m_listLinePPLs)
				ppl.Clear();
		}
	}

	[StructLayout(LayoutKind.Sequential)]
	public struct RadarError
	{
		public uint TSF;
		public byte RSSI;
		public byte Width;
		public byte Type;
		public byte SubType;
	}
}
