﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Diagnostics;

using TehLib;
using Radar.Properties;

namespace Radar
{
	public class MessageList : EventArgs
	{
		public List<string> List = new List<string>();
	}

	public class RadarConnection : ClientConnection
	{
		public RadarConnection()
		{
		}

		public const int RadarPort = 5256;

		public new bool Connect(string strAddress)
		{
			m_strLeftOver = string.Empty;

			return base.Connect(strAddress, RadarPort);
		}

		protected override void OnConnected(StateObject state)
		{
			//
		}

		protected override void OnDisconnected(ConnectionState.Triggers trigger, System.Net.Sockets.SocketError errSocket)
		{
			//
		}

		public bool Send(string str)
		{
			if (Connected == false)
				return false;

			try
			{
				str = str.Trim() + "\n";

				char[] watch = str.ToCharArray();
				byte[] w = new byte[str.Length];
				int used, bytesused;
				bool completed;
				Encoder e = ASCIIEncoding.ASCII.GetEncoder();
				e.Convert(watch, 0, str.Length, w, 0, str.Length, true, out used, out bytesused, out completed);

				Socket.Send(w);
			}
			catch (Exception ex)
			{
				Debug.WriteLine(string.Format("Exception while sending command: {0}", ex));

				return false;
			}

			return true;
		}

		public event EventHandler<MessageList> OnNewMessages = delegate { };

		delegate void ReceiveHandler(string s);

		protected override void OnReceive(StateObject state, int iRead)
		{
			string s = ASCIIEncoding.ASCII.GetString(state.Buffer, 0, iRead);

			UpdateReceiveHandler(s);
		}

		string m_strLeftOver = string.Empty;

		void UpdateReceiveHandler(string strUpdate)
		{
			try
			{
				//if (string.IsNullOrEmpty(m_strLeftOver) == false)
				//	Trace.WriteLine(string.Format("Using leftover: \"{0}\"", m_strLeftOver));

				strUpdate = m_strLeftOver + strUpdate;

				if (strUpdate.Substring(strUpdate.Length - 1) != "\n")
				{
					int iIndex = strUpdate.LastIndexOf('\n');
					m_strLeftOver = strUpdate.Substring(iIndex + 1);
					strUpdate = strUpdate.Substring(0, iIndex + 1);
				}
				else
				{
					//if (string.IsNullOrEmpty(m_strLeftOver) == false)
					//	Debug.WriteLine(string.Format("Clearing leftover: \"{0}\"", m_strLeftOver));

					m_strLeftOver = string.Empty;
				}

				string[] parts = strUpdate.Split('\n');
				MessageList list = new MessageList();
				bool bError = false;

				foreach (string _str in parts)
				{
					string str = _str.Trim();	// '\r'

					if (str.Length == 0)
						continue;

					if (bError)
					{
						Trace.WriteLine(string.Format("Continuing to process data after error: \"{0}\"", str));
						bError = false;
					}

					try
					{
						list.List.Add(str);
					}
					catch (Exception e)
					{
						bError = true;

						Trace.WriteLine(string.Format("Error constructing message: \"{0}\" [Leftover: \"{1}\"] ({2})", str, m_strLeftOver, e.Message));

						m_strLeftOver = string.Empty;
					}
				}

				if (list.List.Count > 0)
				{
					lock (this.DispatchLock)
					{
						if ((this.DispatchForm == null) || (this.DispatchForm.IsDisposed) || (this.DispatchForm.InvokeRequired == false))
							OnNewMessages(this, list);
						else
						{
							//this.DispatchForm.Invoke(OnNewMessage, this, msg);	// Can cause deadlock
							IAsyncResult ar = this.DispatchForm.BeginInvoke(OnNewMessages, this, list);
						}
					}
				}
			}
			catch (Exception e)
			{
				Trace.WriteLine(string.Format("Error handling incoming data: \"{0}\" [Leftover: \"{1}\"] ({2})", strUpdate, m_strLeftOver, e.Message));

				m_strLeftOver = string.Empty;
			}
		}
	}
}
