﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Radar
{
	public class Statistics
	{
		double m_dSum;
		double m_dSumSquared;
		uint m_nCount;
		double? m_dMin, m_dMax;
		Dictionary<double, uint> m_dict;

		public Statistics()
		{
		}

		public Statistics(bool bHisto)
		{
			if (bHisto)
				m_dict = new Dictionary<double, uint>();
		}

		public uint Add(double d)
		{
			m_dSum += d;
			m_dSumSquared += (d * d);

			if (m_dMin.HasValue)
			{
				if (d < m_dMin.Value)
					m_dMin = d;
			}
			else
				m_dMin = d;

			if (m_dMax.HasValue)
			{
				if (d > m_dMax.Value)
					m_dMax = d;
			}
			else
				m_dMax = d;

			if (m_dict != null)
			{
				if (m_dict.ContainsKey(Math.Round(d)))
					m_dict[d] += 1;
				else
					m_dict[d] = 1;
			}

			return ++m_nCount;
		}

		public void Reset()
		{
			m_dSum = 0;
			m_dSumSquared = 0;
			m_nCount = 0;
		}

		private static int CompareKD(KeyValuePair<double, uint> x, KeyValuePair<double, uint> y)
		{
			return (int)y.Key - (int)x.Key;	// Descending
		}

		private static int CompareKA(KeyValuePair<double, uint> x, KeyValuePair<double, uint> y)
		{
			return (int)x.Key - (int)y.Key;	// Ascending
		}

		private static int CompareVD(KeyValuePair<double, uint> x, KeyValuePair<double, uint> y)
		{
			return (int)y.Value - (int)x.Value;	// Descending
		}

		private static int CompareVA(KeyValuePair<double, uint> x, KeyValuePair<double, uint> y)
		{
			return (int)x.Value - (int)y.Value;	// Ascending
		}

		public enum HistogramSorting
		{
			None,
			BinAsc,
			BinDesc,
			CountAsc,
			CountDesc
		}

		public List<KeyValuePair<double, uint>> GetHistogram(HistogramSorting hs)
		{
			if (m_dict == null)
				return new  List<KeyValuePair<double, uint>>();

			List<KeyValuePair<double, uint>> list = m_dict.ToList();
			switch (hs)
			{
				case HistogramSorting.BinAsc:
					list.Sort(CompareKA);
					break;
				case HistogramSorting.BinDesc:
					list.Sort(CompareKD);
					break;
				case HistogramSorting.CountAsc:
					list.Sort(CompareVA);
					break;
				case HistogramSorting.CountDesc:
					list.Sort(CompareVD);
					break;
			}

			return list;
		}

		public double Min
		{
			get { return m_dMin.GetValueOrDefault(0); }
		}

		public double Max
		{
			get { return m_dMax.GetValueOrDefault(0); }
		}

		public double Average
		{
			get
			{
				if (m_nCount == 0)
					return 0;
				return (m_dSum / (double)m_nCount);
			}
		}

		public double Variance
		{
			get
			{
				if (m_nCount == 0)
					return 0;
				return (m_dSumSquared / (double)m_nCount) - Math.Pow(Average, 2.0);
			}
		}

		public double StandardDeviation
		{
			get { return Math.Sqrt(Variance); }
		}
	}
}
